from tensorflow import keras
from teras.layers.tabnet.tabnet_encoder import TabNetEncoder
from teras.layers.tabnet.tabnet_decoder import TabNetDecoder
from teras.layers.categorical_feature_embedding import CategoricalFeatureEmbedding
from teras.layers.common.head import (ClassificationHead,
                                      RegressionHead)
from teras.config.tabnet import TabNetConfig
from teras.layerflow.models.tabnet import (TabNet as _TabNetLF,
                                           TabNetPretrainer as _TabNetPretrainerLF)
from teras.utils.types import UnitsValuesType


@keras.saving.register_keras_serializable(package="keras.models")
class TabNet(_TabNetLF):
    """
    TabNet model as proposed by Sercan et al. in TabNet paper.
    This purpose will serve as the parent class for the TabNetRegressor and TabNetClassifier.

    Reference(s):
        https://arxiv.org/abs/1908.07442

    Args:
        input_dim: ``int``,
            Dimensionality of the input dataset,
            or the number of features in the input dataset.

        features_metadata: ``dict``,
            A nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            ``{feature_name: (feature_idx, vocabulary)}`` for feature in categorical features.
            ``{feature_name: feature_idx}`` for feature in numerical features.
            You can get this dictionary from
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> metadata_dict = get_features_metadata_for_embedding(dataframe,
                ..                                                      categorical_features,
                ..                                                      numerical_features)

        feature_transformer_dim: ``int``, default 32,
            The dimensionality of the hidden representation in feature transformation block.
            Each layer first maps the representation to a `2 * feature_transformer_dim`
            output and half of it is used to determine the
            non-linearity of the GLU activation where the other half is used as an
            input to GLU, and eventually `feature_transformer_dim` output is
            transferred to the next layer.

        decision_step_output_dim: ``int``, default 32,
            The dimensionality of output at each decision step, which is later mapped to the
            final classification or regression output.
            It is recommended to keep ``decision_step_output_dim`` and ``feature_transformer_dim``
            equal to each other.
            Adjusting these two parameters values is a good way of obtaining a tradeoff between
            performance and complexity.

        num_decision_steps: ``int``, default 5,
            The number of sequential decision steps.
            For most datasets a value in the range [3, 10] is optimal.
            If there are more informative features in the dataset, the value tends to
            be higher. That said, a very high value of `num_decision_steps` may suffer
            from overfitting.

        num_shared_layers: ``int``, default 2,
            Number of shared layers to use in the ``TabNetFeatureTransformer``.
            These shared layers are *shared* across decision steps.

        num_decision_dependent_layers: ``int``, default 2,
            Number of decision dependent layers to use in the ``TabNetFeatureTransformer``.
            In simple words, ``num_decision_dependent_layers`` are created
            for each decision step in the ``num_decision_steps``.
            For instance, if ``num_decision_steps = `5` and  ``num_decision_dependent_layers = 2``
            then 10 layers will be created, 2 for each decision step.

        relaxation_factor: ``float``, default 1.5,
            Relaxation factor that promotes the reuse of each
            feature at different decision steps. When it is 1, a feature is enforced
            to be used only at one decision step and as it increases, more
            flexibility is provided to use a feature at multiple decision steps.
            An optimal value of relaxation_factor can have a major role on the performance.
            Typically, a larger value for `num_decision_steps` favors for a larger ``relaxation_factor``.

        batch_momentum: ``float``, default 0.9,
            Momentum value to use for ``BatchNormalization`` layer.

        virtual_batch_size: `int`, default 64,
            Batch size to use for ``virtual_batch_size`` parameter in ``BatchNormalization`` layer.
            This is typically much smaller than the ``batch_size`` used for training.

        residual_normalization_factor: ``float``, default 0.5,
            In the feature transformer, except for the first layer, every other layer utilizes
            normalized residuals, where ``residual_normalization_factor``
            determines the scale of normalization.

        epsilon: ``float``, default 0.00001,
            Epsilon is a small number for numerical stability
            during the computation of entropy loss.

        encode_categorical_values: ``bool``, default True,
            Whether to (label) encode categorical values,
            If you've already encoded the categorical values using for instance
            Label/Ordinal encoding, you should set this to False,
            otherwise leave it as True.
            In the case of True, categorical values will be mapped to integer indices
            using keras's ``IntegerLookup`` layer.
    """
    def __init__(self,
                 input_dim: int,
                 features_metadata: dict,
                 feature_transformer_dim: int = TabNetConfig.feature_transformer_dim,
                 decision_step_output_dim: int = TabNetConfig.decision_step_output_dim,
                 num_decision_steps: int = TabNetConfig.num_decision_steps,
                 num_shared_layers: int = TabNetConfig.num_shared_layers,
                 num_decision_dependent_layers: int = TabNetConfig.num_decision_dependent_layers,
                 relaxation_factor: float = TabNetConfig.relaxation_factor,
                 batch_momentum: float = TabNetConfig.batch_momentum,
                 virtual_batch_size: int = TabNetConfig.virtual_batch_size,
                 residual_normalization_factor: float = TabNetConfig.residual_normalization_factor,
                 epsilon: float = TabNetConfig.epsilon,
                 encode_categorical_values: bool = TabNetConfig.encode_categorical_features,
                 **kwargs):
        categorical_feature_embedding = CategoricalFeatureEmbedding(features_metadata=features_metadata,
                                                                     embedding_dim=1,
                                                                     encode=encode_categorical_values)
        encoder = TabNetEncoder(data_dim=input_dim,
                                feature_transformer_dim=feature_transformer_dim,
                                decision_step_output_dim=decision_step_output_dim,
                                num_decision_steps=num_decision_steps,
                                num_shared_layers=num_shared_layers,
                                num_decision_dependent_layers=num_decision_dependent_layers,
                                relaxation_factor=relaxation_factor,
                                batch_momentum=batch_momentum,
                                virtual_batch_size=virtual_batch_size,
                                residual_normalization_factor=residual_normalization_factor,
                                epsilon=epsilon)
        super().__init__(input_dim=input_dim,
                         features_metadata=features_metadata,
                         categorical_feature_embedding=categorical_feature_embedding,
                         encoder=encoder,
                         **kwargs)
        self.input_dim = input_dim
        self.features_metadata = features_metadata
        self.feature_transformer_dim = feature_transformer_dim
        self.decision_step_output_dim = decision_step_output_dim
        self.num_decision_steps = num_decision_steps
        self.num_shared_layers = num_shared_layers
        self.num_decision_dependent_layers = num_decision_dependent_layers
        self.relaxation_factor = relaxation_factor
        self.batch_momentum = batch_momentum
        self.virtual_batch_size = virtual_batch_size
        self.residual_normalization_factor = residual_normalization_factor
        self.epsilon = epsilon
        self.encode_categorical_values = encode_categorical_values

    def get_config(self):
        config = {'name': self.name,
                  'trainable': self.trainable,
                  'input_dim': self.input_dim,
                  'features_metadata': self.features_metadata,
                  'feature_transformer_dim': self.feature_transformer_dim,
                  'decision_step_output_dim': self.decision_step_output_dim,
                  'num_decision_steps': self.num_decision_steps,
                  'num_shared_layers': self.num_shared_layers,
                  'num_decision_dependent_layers': self.num_decision_dependent_layers,
                  'relaxation_factor': self.relaxation_factor,
                  'batch_momentum': self.batch_momentum,
                  'virtual_batch_size': self.virtual_batch_size,
                  'residual_normalization_factor': self.residual_normalization_factor,
                  'epsilon': self.epsilon,
                  'encode_categorical_values': self.encode_categorical_values,
                  }
        return config

    @classmethod
    def from_config(cls, config):
        input_dim = config.pop("input_dim")
        features_metadata = config.pop("features_metadata")
        return cls(input_dim=input_dim,
                   features_metadata=features_metadata,
                   **config)


@keras.saving.register_keras_serializable(package="keras.models")
class TabNetClassifier(TabNet):
    """
    TabNet Classifier based on the TabNet architecture
    as proposed by Sercan et al. in TabNet paper.

    Reference(s):
        https://arxiv.org/abs/1908.07442

    Args:
        num_classes: ``int``, default 2,
            Number of classes to predict.

        head_units_values: ``List[int]`` or ``Tuple[int]``, default None,
            Hidden units to use in the Classification head.
            For each value in the list/tuple,
            a hidden layer of that dimensionality is added to the head.
            By default, no hidden layer is used.

        input_dim: ``int``,
            Dimensionality of the input dataset,
            or the number of features in the input dataset.

        features_metadata: ``dict``,
            A nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            ``{feature_name: (feature_idx, vocabulary)}`` for feature in categorical features.
            ``{feature_name: feature_idx}`` for feature in numerical features.
            You can get this dictionary from
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> metadata_dict = get_features_metadata_for_embedding(dataframe,
                ..                                                      categorical_features,
                ..                                                      numerical_features)

        feature_transformer_dim: ``int``, default 32,
            The dimensionality of the hidden representation in feature transformation block.
            Each layer first maps the representation to a `2 * feature_transformer_dim`
            output and half of it is used to determine the
            non-linearity of the GLU activation where the other half is used as an
            input to GLU, and eventually `feature_transformer_dim` output is
            transferred to the next layer.

        decision_step_output_dim: ``int``, default 32,
            The dimensionality of output at each decision step, which is later mapped to the
            final classification or regression output.
            It is recommended to keep ``decision_step_output_dim`` and ``feature_transformer_dim``
            equal to each other.
            Adjusting these two parameters values is a good way of obtaining a tradeoff between
            performance and complexity.

        num_decision_steps: ``int``, default 5,
            The number of sequential decision steps.
            For most datasets a value in the range [3, 10] is optimal.
            If there are more informative features in the dataset, the value tends to
            be higher. That said, a very high value of `num_decision_steps` may suffer
            from overfitting.

        num_shared_layers: ``int``, default 2,
            Number of shared layers to use in the ``TabNetFeatureTransformer``.
            These shared layers are *shared* across decision steps.

        num_decision_dependent_layers: ``int``, default 2,
            Number of decision dependent layers to use in the ``TabNetFeatureTransformer``.
            In simple words, ``num_decision_dependent_layers`` are created
            for each decision step in the ``num_decision_steps``.
            For instance, if ``num_decision_steps = `5` and  ``num_decision_dependent_layers = 2``
            then 10 layers will be created, 2 for each decision step.

        relaxation_factor: ``float``, default 1.5,
            Relaxation factor that promotes the reuse of each
            feature at different decision steps. When it is 1, a feature is enforced
            to be used only at one decision step and as it increases, more
            flexibility is provided to use a feature at multiple decision steps.
            An optimal value of relaxation_factor can have a major role on the performance.
            Typically, a larger value for `num_decision_steps` favors for a larger ``relaxation_factor``.

        batch_momentum: ``float``, default 0.9,
            Momentum value to use for ``BatchNormalization`` layer.

        virtual_batch_size: `int`, default 64,
            Batch size to use for ``virtual_batch_size`` parameter in ``BatchNormalization`` layer.
            This is typically much smaller than the ``batch_size`` used for training.

        residual_normalization_factor: ``float``, default 0.5,
            In the feature transformer, except for the first layer, every other layer utilizes
            normalized residuals, where ``residual_normalization_factor``
            determines the scale of normalization.

        epsilon: ``float``, default 0.00001,
            Epsilon is a small number for numerical stability
            during the computation of entropy loss.

        encode_categorical_values: ``bool``, default True,
            Whether to (label) encode categorical values,
            If you've already encoded the categorical values using for instance
            Label/Ordinal encoding, you should set this to False,
            otherwise leave it as True.
            In the case of True, categorical values will be mapped to integer indices
            using keras's ``IntegerLookup`` layer.
    """
    def __init__(self,
                 num_classes: int = 2,
                 head_units_values: UnitsValuesType = None,
                 input_dim: int = None,
                 features_metadata: dict = None,
                 feature_transformer_dim: int = TabNetConfig.feature_transformer_dim,
                 decision_step_output_dim: int = TabNetConfig.decision_step_output_dim,
                 num_decision_steps: int = TabNetConfig.num_decision_steps,
                 num_shared_layers: int = TabNetConfig.num_shared_layers,
                 num_decision_dependent_layers: int = TabNetConfig.num_decision_dependent_layers,
                 relaxation_factor: float = TabNetConfig.relaxation_factor,
                 batch_momentum: float = TabNetConfig.batch_momentum,
                 virtual_batch_size: int = TabNetConfig.virtual_batch_size,
                 residual_normalization_factor: float = TabNetConfig.residual_normalization_factor,
                 epsilon: float = TabNetConfig.epsilon,
                 encode_categorical_values: bool = TabNetConfig.encode_categorical_features,
                 **kwargs):
        head = ClassificationHead(num_classes=num_classes,
                                  units_values=head_units_values,
                                  name="tabnet_classification_head")
        super().__init__(input_dim=input_dim,
                         features_metadata=features_metadata,
                         feature_transformer_dim=feature_transformer_dim,
                         decision_step_output_dim=decision_step_output_dim,
                         num_decision_steps=num_decision_steps,
                         num_shared_layers=num_shared_layers,
                         num_decision_dependent_layers=num_decision_dependent_layers,
                         relaxation_factor=relaxation_factor,
                         batch_momentum=batch_momentum,
                         virtual_batch_size=virtual_batch_size,
                         residual_normalization_factor=residual_normalization_factor,
                         epsilon=epsilon,
                         encode_categorical_values=encode_categorical_values,
                         head=head,
                         **kwargs)
        self.num_classes = num_classes
        self.head_units_values = head_units_values

    def get_config(self):
        config = super().get_config()
        config.update({'num_classes': self.num_classes,
                       'head_units_values': self.head_units_values
                       })
        return config


@keras.saving.register_keras_serializable(package="keras.models")
class TabNetRegressor(TabNet):
    """
    TabNet Regressor based on the TabNet architecture
    as proposed by Sercan et al. in TabNet paper.

    Reference(s):
        https://arxiv.org/abs/1908.07442

    Args:
        num_outputs: ``int``, default 1,
            Number of regression outputs.

        head_units_values: ``List[int]`` or ``Tuple[int]``, default None,
            Hidden units to use in the Classification head.
            For each value in the list/tuple,
            a hidden layer of that dimensionality is added to the head.
            By default, no hidden layer is used.

        input_dim: ``int``,
            Dimensionality of the input dataset,
            or the number of features in the input dataset.

        features_metadata: ``dict``,
            A nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            ``{feature_name: (feature_idx, vocabulary)}`` for feature in categorical features.
            ``{feature_name: feature_idx}`` for feature in numerical features.
            You can get this dictionary from
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> metadata_dict = get_features_metadata_for_embedding(dataframe,
                ..                                                      categorical_features,
                ..                                                      numerical_features)

        feature_transformer_dim: ``int``, default 32,
            The dimensionality of the hidden representation in feature transformation block.
            Each layer first maps the representation to a `2 * feature_transformer_dim`
            output and half of it is used to determine the
            non-linearity of the GLU activation where the other half is used as an
            input to GLU, and eventually `feature_transformer_dim` output is
            transferred to the next layer.

        decision_step_output_dim: ``int``, default 32,
            The dimensionality of output at each decision step, which is later mapped to the
            final classification or regression output.
            It is recommended to keep ``decision_step_output_dim`` and ``feature_transformer_dim``
            equal to each other.
            Adjusting these two parameters values is a good way of obtaining a tradeoff between
            performance and complexity.

        num_decision_steps: ``int``, default 5,
            The number of sequential decision steps.
            For most datasets a value in the range [3, 10] is optimal.
            If there are more informative features in the dataset, the value tends to
            be higher. That said, a very high value of `num_decision_steps` may suffer
            from overfitting.

        num_shared_layers: ``int``, default 2,
            Number of shared layers to use in the ``TabNetFeatureTransformer``.
            These shared layers are *shared* across decision steps.

        num_decision_dependent_layers: ``int``, default 2,
            Number of decision dependent layers to use in the ``TabNetFeatureTransformer``.
            In simple words, ``num_decision_dependent_layers`` are created
            for each decision step in the ``num_decision_steps``.
            For instance, if ``num_decision_steps = `5` and  ``num_decision_dependent_layers = 2``
            then 10 layers will be created, 2 for each decision step.

        relaxation_factor: ``float``, default 1.5,
            Relaxation factor that promotes the reuse of each
            feature at different decision steps. When it is 1, a feature is enforced
            to be used only at one decision step and as it increases, more
            flexibility is provided to use a feature at multiple decision steps.
            An optimal value of relaxation_factor can have a major role on the performance.
            Typically, a larger value for `num_decision_steps` favors for a larger ``relaxation_factor``.

        batch_momentum: ``float``, default 0.9,
            Momentum value to use for ``BatchNormalization`` layer.

        virtual_batch_size: `int`, default 64,
            Batch size to use for ``virtual_batch_size`` parameter in ``BatchNormalization`` layer.
            This is typically much smaller than the ``batch_size`` used for training.

        residual_normalization_factor: ``float``, default 0.5,
            In the feature transformer, except for the first layer, every other layer utilizes
            normalized residuals, where ``residual_normalization_factor``
            determines the scale of normalization.

        epsilon: ``float``, default 0.00001,
            Epsilon is a small number for numerical stability
            during the computation of entropy loss.

        encode_categorical_values: ``bool``, default True,
            Whether to (label) encode categorical values,
            If you've already encoded the categorical values using for instance
            Label/Ordinal encoding, you should set this to False,
            otherwise leave it as True.
            In the case of True, categorical values will be mapped to integer indices
            using keras's ``IntegerLookup`` layer.
    """
    def __init__(self,
                 num_outputs=1,
                 head_units_values: UnitsValuesType = None,
                 input_dim: int = None,
                 features_metadata: dict = None,
                 feature_transformer_dim: int = TabNetConfig.feature_transformer_dim,
                 decision_step_output_dim: int = TabNetConfig.decision_step_output_dim,
                 num_decision_steps: int = TabNetConfig.num_decision_steps,
                 num_shared_layers: int = TabNetConfig.num_shared_layers,
                 num_decision_dependent_layers: int = TabNetConfig.num_decision_dependent_layers,
                 relaxation_factor: float = TabNetConfig.relaxation_factor,
                 batch_momentum: float = TabNetConfig.batch_momentum,
                 virtual_batch_size: int = TabNetConfig.virtual_batch_size,
                 residual_normalization_factor: float = TabNetConfig.residual_normalization_factor,
                 epsilon: float = TabNetConfig.epsilon,
                 encode_categorical_values: bool = TabNetConfig.encode_categorical_features,
                 **kwargs):
        head = RegressionHead(num_outputs=num_outputs,
                              units_values=head_units_values,
                              name="tabnet_regression_head")
        super().__init__(input_dim=input_dim,
                         features_metadata=features_metadata,
                         feature_transformer_dim=feature_transformer_dim,
                         decision_step_output_dim=decision_step_output_dim,
                         num_decision_steps=num_decision_steps,
                         num_shared_layers=num_shared_layers,
                         num_decision_dependent_layers=num_decision_dependent_layers,
                         relaxation_factor=relaxation_factor,
                         batch_momentum=batch_momentum,
                         virtual_batch_size=virtual_batch_size,
                         residual_normalization_factor=residual_normalization_factor,
                         epsilon=epsilon,
                         encode_categorical_values=encode_categorical_values,
                         head=head,
                         **kwargs)
        self.num_outputs = num_outputs
        self.head_units_values = head_units_values

    def get_config(self):
        config = super().get_config()
        config.update({'num_outputs': self.num_outputs,
                       'head_units_values': self.head_units_values
                       })
        return config


@keras.saving.register_keras_serializable(package="keras.models")
class TabNetPretrainer(_TabNetPretrainerLF):
    """
    TabNetPretrainer model based on the architecture
    proposed by Sercan et al. in TabNet paper.

    TabNetPretrainer is an encoder-decoder model based on the TabNet architecture,
    where the TabNet model acts as an encoder while a separate decoder
    is used to reconstruct the input features.

    Reference(s):
        https://arxiv.org/abs/1908.07442

    Args:
        model: ``TabNet``,
            An instance of `TabNet` class to pretrain.

        input_dim: ``int``,
            Dimensionality of the input dataset,
            or the number of features in the input dataset.

        features_metadata: ``dict``,
            A nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            ``{feature_name: (feature_idx, vocabulary)}`` for feature in categorical features.
            ``{feature_name: feature_idx}`` for feature in numerical features.
            You can get this dictionary from
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> metadata_dict = get_features_metadata_for_embedding(dataframe,
                ..                                                      categorical_features,
                ..                                                      numerical_features)

        missing_feature_probability: ``float``, default 3,
            Fraction of features to randomly mask i.e. make them missing.
            Missing features are introduced in the pretraining dataset and
            the probability of missing features is controlled by the parameter.
            The pretraining objective is to predict values for these missing features,
            (pre)training the ``TabNet`` model in the process.

        decoder_feature_transformer_dim: ``int``, default 32,
            Feature transformer dimensions for ``TabNetDecoder``.

        decoder_decision_step_output_dim: ``int``, default 32,
            Decision step output dimensions for ``TabNetDecoder``.

        decoder_num_decision_steps: ``int``, default 5,
            Number of decision steps to use in ``TabNetDecoder``.

        decoder_num_shared_layers: ``int``, default 2,
            Number of shared layers in ``TabNetFeatureTransformer`` in ``TabNetDecoder``.

        decoder_num_decision_dependent_layers: ``int``, default 2,
            Number of decision dependent layers in ``TabNetFeatureTransformer`` layer in ``TabNetDecoder``.
    """
    def __init__(self,
                 model: TabNet,
                 input_dim: int = None,
                 features_metadata: dict = None,
                 missing_feature_probability: float = 0.3,
                 decoder_feature_transformer_dim: int = TabNetConfig.feature_transformer_dim,
                 decoder_decision_step_output_dim: int = TabNetConfig.decision_step_output_dim,
                 decoder_num_decision_steps: int = TabNetConfig.num_decision_steps,
                 decoder_num_shared_layers: int = TabNetConfig.num_shared_layers,
                 decoder_num_decision_dependent_layers: int = TabNetConfig.num_decision_dependent_layers,
                 **kwargs):
        decoder = TabNetDecoder(data_dim=input_dim,
                                feature_transformer_dim=decoder_feature_transformer_dim,
                                decision_step_output_dim=decoder_decision_step_output_dim,
                                num_decision_steps=decoder_num_decision_steps,
                                num_shared_layers=decoder_num_shared_layers,
                                num_decision_dependent_layers=decoder_num_shared_layers)
        super().__init__(model=model,
                         features_metadata=features_metadata,
                         decoder=decoder,
                         missing_feature_probability=missing_feature_probability,
                         **kwargs)
        self.model = model
        self.input_dim = input_dim
        self.features_metadata = features_metadata
        self.missing_feature_probability = missing_feature_probability
        self.decoder_feature_transformer_dim = decoder_feature_transformer_dim
        self.decoder_decision_step_output_dim = decoder_decision_step_output_dim
        self.decoder_num_decision_steps = decoder_num_decision_steps
        self.decoder_num_shared_layers = decoder_num_shared_layers
        self.decoder_num_decision_dependent_layers = decoder_num_decision_dependent_layers

    def get_config(self):
        config = {'name': self.name,
                  'trainable': self.trainable,
                  'model': keras.layers.serialize(self.model),
                  'input_dim': self.input_dim,
                  'features_metadata': self.features_metadata,
                  'missing_feature_probability': self.missing_feature_probability,
                  'decoder_feature_transformer_dim': self.decoder_feature_transformer_dim,
                  'decoder_decision_step_output_dim': self.decoder_decision_step_output_dim,
                  'decoder_num_decision_steps': self.decoder_num_decision_steps,
                  'decoder_num_shared_layers': self.decoder_num_shared_layers,
                  'decoder_num_decision_dependent_layers': self.decoder_num_decision_dependent_layers,
                  }
        return config

    @classmethod
    def from_config(cls, config):
        model = keras.layers.deserialize(config.pop("model"))
        return cls(model=model, **config)



from tensorflow import keras
from teras.layers.gain import GAINGeneratorBlock, GAINDiscriminatorBlock
from teras.layerflow.models.gain import GAIN as _GAIN_LF
from teras.utils.types import UnitsValuesType, ActivationType


@keras.saving.register_keras_serializable(package="teras.models")
class GAINGenerator(keras.Model):
    """
    Generator model for the GAIN architecture proposed by
    Jinsung Yoon et al. in the paper
    GAIN: Missing Data Imputation using Generative Adversarial Nets.

    Reference(s):
        https://arxiv.org/abs/1806.02920

    Args:
        data_dim: ``int``,
            The dimensionality of the input dataset.

            Note the dimensionality must be equal to the dimensionality of dataset
            that is passed to the fit method and not necessarily the dimensionality
            of the raw input dataset as sometimes data transformation alters the
            dimensionality of the dataset.

            You can access the dimensionality of the transformed dataset through the
            ``.data_dim`` attribute of the ``GAINDataSampler`` instance used in sampling
            the dataset.

        units_values: ``List[int]`` or ``Tuple[int]``,
            A list or tuple of units for constructing hidden block.
            For each value, a ``GAINGeneratorBlock``
            (``from teras.layers import GAINGeneratorBlock``)
            of that dimensionality (units) is added to the generator
            to form the ``hidden block`` of the generator.
            By default, ``units_values`` = [``data_dim``, ``data_dim``].

        activation_hidden: default "relu",
            Activation function to use for the hidden layers in the hidden block.

        activation_out: default "sigmoid",
            Activation function to use for the output layer of the Generator.

    Example:
        ```python
        data_dim = 12
        # Instantiate Generator.
        generator = Generator(data_dim=data_dim)

        # Sample noise to generate samples from
        z = tf.random.normal([512, 18])

        # Generate samples
        generated_samples = generator(z)
        ```
    """
    def __init__(self,
                 data_dim: int,
                 units_values: UnitsValuesType = None,
                 activation_hidden: ActivationType = "relu",
                 activation_out: ActivationType = "sigmoid",
                 **kwargs):
        super().__init__(**kwargs)

        if units_values is not None and not isinstance(units_values, (list, tuple)):
            raise ValueError(f"""`units_values` must be a list or tuple of units which determines
                        the number of Generator blocks and the dimensionality of those blocks.
                        But received type {type(units_values)} which is not supported.""")

        self.data_dim = data_dim
        self.units_values = units_values
        self.activation_hidden = activation_hidden
        self.activation_out = activation_out

        if self.units_values is None:
            self.units_values = [self.data_dim] * 2
        self.hidden_block = keras.models.Sequential(name="generator_hidden_block")
        for units in self.units_values:
            self.hidden_block.add(GAINGeneratorBlock(units,
                                                     activation=self.activation_hidden))

        self.output_layer = keras.layers.Dense(self.data_dim,
                                               activation=self.activation_out,
                                               name="generator_output_layer")

    def call(self, inputs):
        # inputs is the concatenation of `mask` and `original data`
        # where `mask` has the same dimensions as data
        # so the inputs received are 2x the dimensions of `original data`
        outputs = self.hidden_block(inputs)
        outputs = self.output_layer(outputs)
        return outputs

    def get_config(self):
        config = super().get_config()
        activation_hidden_serialized = self.activation_hidden
        if not isinstance(self.activation_hidden, str):
            activation_hidden_serialized = keras.layers.serialize(self.activation_hidden)

        activation_out_serialized = self.activation_out
        if not isinstance(self.activation_out, str):
            activation_out_serialized = keras.layers.serialize(self.activation_out)

        config.update({'data_dim': self.data_dim,
                       "units_values": self.units_values,
                       "activation_hidden": activation_hidden_serialized,
                       "activation_out": activation_out_serialized,
                       }
                      )
        return config

    @classmethod
    def from_config(cls, config):
        data_dim = config.pop("data_dim")
        return cls(data_dim=data_dim,
                   **config)


@keras.saving.register_keras_serializable(package="teras.models")
class GAINDiscriminator(keras.Model):
    """
    Discriminator model for the GAIN architecture proposed by
    Jinsung Yoon et al. in the paper
    GAIN: Missing Data Imputation using Generative Adversarial Nets.

    Note that the Generator and Discriminator share the exact same
    architecture by default. They differ in the inputs they receive
    and their loss functions.

    Reference(s):
        https://arxiv.org/abs/1806.02920

    Args:
        data_dim: ``int``,
            The dimensionality of the input dataset.

            Note the dimensionality must be equal to the dimensionality of dataset
            that is passed to the fit method and not necessarily the dimensionality
            of the raw input dataset as sometimes data transformation alters the
            dimensionality of the dataset.

            You can access the dimensionality of the transformed dataset through the
            ``.data_dim`` attribute of the ``GAINDataSampler`` instance used in sampling
            the dataset.

        units_values: ``List[int]`` or ``Tuple[int]``,
            A list or tuple of units for constructing hidden block.
            For each value, a ``GAINDiscriminatorBlock``
            (``from teras.layers import GAINDiscriminatorBlock``)
            of that dimensionality (units) is added to the discriminator
            to form the ``hidden block`` of the discriminator.
            By default, ``units_values`` = [``data_dim``, ``data_dim``].

        activation_hidden: default "relu",
            Activation function to use for the hidden layers in the hidden block.

        activation_out: default "sigmoid",
            Activation function to use for the output layer of the Discriminator.

    Example:
        ```python
        data_dim = 12
        # Instantiate Generator
        generator = Generator(data_dim)

        # Instantiate Discriminator
        discriminator = Discriminator(data_dim)

        # Sample noise to generate samples from
        z = tf.random.normal([512, 18])

        # Generate samples
        generated_samples = generator(z)

        # Predict using discriminator
        y_pred = discriminator(generated_samples)
        ```
    """
    def __init__(self,
                 data_dim: int,
                 units_values: UnitsValuesType = None,
                 activation_hidden: ActivationType = "relu",
                 activation_out: ActivationType = "sigmoid",
                 **kwargs):
        super().__init__(**kwargs)

        if units_values is not None and not isinstance(units_values, (list, tuple)):
            raise ValueError(f"""`units_values` must be a list or tuple of units which determines
                        the number of Discriminator blocks and the dimensionality of those blocks.
                        But {units_values} was passed.""")

        self.data_dim = data_dim
        self.units_values = units_values
        self.activation_hidden = activation_hidden
        self.activation_out = activation_out

        if self.units_values is None:
            self.units_values = [self.data_dim] * 2
        self.hidden_block = keras.models.Sequential(name="discriminator_hidden_block")
        for units in self.units_values:
            self.hidden_block.add(GAINDiscriminatorBlock(units,
                                                         activation=self.activation_hidden))

        self.output_layer = keras.layers.Dense(self.data_dim,
                                               activation=self.activation_out,
                                               name="discriminator_output_layer")

    def call(self, inputs):
        # inputs is the concatenation of `hint` and manipulated
        # Generator output (i.e. generated samples).
        # `hint` has the same dimensions as data
        # so the inputs received are 2x the dimensions of original data
        outputs = self.hidden_block(inputs)
        outputs = self.output_layer(outputs)
        return outputs

    def get_config(self):
        config = super().get_config()
        activation_hidden_serialized = self.activation_hidden
        if not isinstance(self.activation_hidden, str):
            activation_hidden_serialized = keras.layers.serialize(self.activation_hidden)

        activation_out_serialized = self.activation_out
        if not isinstance(self.activation_out, str):
            activation_out_serialized = keras.layers.serialize(self.activation_out)

        config.update({'data_dim': self.data_dim,
                       "units_values": self.units_values,
                       "activation_hidden": activation_hidden_serialized,
                       "activation_out": activation_out_serialized,
                       }
                      )
        return config

    @classmethod
    def from_config(cls, config):
        data_dim = config.pop("data_dim")
        return cls(data_dim=data_dim,
                   **config)


@keras.saving.register_keras_serializable(package="teras.models")
class GAIN(_GAIN_LF):
    """
    GAIN is a missing data imputation model based on GANs.
    This is an implementation of the GAIN architecture
    proposed by Jinsung Yoon et al. in the paper
    GAIN: Missing Data Imputation using Generative Adversarial Nets.

    In GAIN, the generator observes some components of
    a real data vector, imputes the missing components
    conditioned on what is actually observed, and
    outputs a completed vector.
    The discriminator then takes a completed vector
    and attempts to determine which components
    were actually observed and which were imputed.
    It also utilizes a novel hint mechanism, which
    ensures that generator does in fact learn to generate
    samples according to the true data distribution.

    Reference(s):
        https://arxiv.org/abs/1806.02920

    Args:
        input_dim: ``int``,
            The dimensionality of the input dataset.

            Note the dimensionality must be equal to the dimensionality of dataset
            that is passed to the fit method and not necessarily the dimensionality
            of the raw input dataset as sometimes data transformation alters the
            dimensionality of the dataset.

            You can access the dimensionality of the transformed dataset through the
            ``.data_dim`` attribute of the ``GAINDataSampler`` instance used in sampling
            the dataset.

        generator_units_values: ``List[int]`` or ``Tuple[int]``,
            A list or tuple of units for constructing hidden block for the Generator.
            For each value, a ``GAINGeneratorBlock``
            (``from teras.layers import GAINGeneratorBlock``)
            of that dimensionality (units) is added to the generator to form the
            `hidden block` of the generator.
            By default, ``generator_units_values = [data_dim, data_dim]``.

        discriminator_units_values: ``List[int]`` or ``Tuple[int]``,
            A list or tuple of units for constructing hidden block for the Discriminator.
            For each value, a ``GAINDiscriminatorBlock``
            (``from teras.layers import GAINDiscriminatorBlock``)
            of that dimensionality (units) is added to the discriminator
            to form the `hidden block` of the discriminator.
            By default, ``discriminator_units_values = [data_dim, data_dim]``.

        generator_activation_hidden: default "relu",
            Activation function to use for the hidden layers in the hidden block
            for the Generator.

        discriminator_activation_hidden: default "relu",
            Activation function to use for the hidden layers in the hidden block
            for the Discriminator.

        generator_activation_out: default "sigmoid",
            Activation function to use for the output layer of the Generator.

        discriminator_activation_out: default "sigmoid",
            Activation function to use for the output layer of the Discriminator.

        num_discriminator_steps: ``int``, default 1,
            Number of discriminator training steps per GAIN training step.

        hint_rate: ``float``, default 0.9,
            Hint rate will be used to sample binary vectors for
            `hint vectors` generation. Must be between 0. and 1.
            Hint vectors ensure that generated samples follow the
            underlying data distribution.

        alpha: ``float``, default 100,
            Hyper parameter for the generator loss computation that
            controls how much weight should be given to the MSE loss.
            Precisely, `generator_loss` = `cross_entropy_loss` + `alpha` * `mse_loss`
            The higher the `alpha`, the more the mse_loss will affect the
            overall generator loss.
    Example:
        ```python
        input_data = pd.DataFrame(tf.random.uniform([200, 5]),
                                columns=["A", "B", "C", "D", "E"])

        from teras.utils.gain import inject_missing_values
        input_data = inject_missing_values(input_data)

        # GAIN requires data in a specific format for which we have
        # relevant DataSampler and DataTrasnformer classes
        from teras.preprocessing.gain import DataSampler, DataTransformer

        numerical_features = ["A", "B"]
        categorical_features = ["C", "D", "E"]

        data_transformer = DataTransformer(numerical_features=numerical_features,
                                           categorical_features=categorical_features)

        transformed_data = data_transformer.transform(input_data, return_dataframe=True)

        data_sampler = DataSampler()
        dataset = data_sampler.get_dataset(transformed_data)

        # Instantiate GAIN
        gain_imputer = GAIN(input_dim=data_sampler.data_dim)

        # Compile it
        gain_imputer.compile()

        # Train it
        gain_imputer.fit(dataset)

        # Predict
        test_data = transformed_data[:50]

        # Imputation Method 1:
        imputed_data = gain_imputer.predict(test_data)

        # Reverse transform into original format
        imputed_data = data_transformer.reverse_transform(imputed_data)

        # Imputation Method 2:
        imputed_data = gain_imputer.impute(test_data)
        ```
    """
    def __init__(self,
                 input_dim: int,
                 generator_units_values: UnitsValuesType = None,
                 discriminator_units_values: UnitsValuesType = None,
                 generator_activation_hidden: ActivationType = "relu",
                 discriminator_activation_hidden: ActivationType = "relu",
                 generator_activation_out: ActivationType = "sigmoid",
                 discriminator_activation_out: ActivationType = "sigmoid",
                 num_discriminator_steps: int = 1,
                 hint_rate: float = 0.9,
                 alpha: float = 100,
                 **kwargs):
        generator = GAINGenerator(data_dim=input_dim,
                                  units_values=generator_units_values,
                                  activation_hidden=generator_activation_hidden,
                                  activation_out=generator_activation_out)
        discriminator = GAINDiscriminator(data_dim=input_dim,
                                          units_values=discriminator_units_values,
                                          activation_hidden=discriminator_activation_hidden,
                                          activation_out=discriminator_activation_out)
        super().__init__(generator=generator,
                         discriminator=discriminator,
                         num_discriminator_steps=num_discriminator_steps,
                         hint_rate=hint_rate,
                         alpha=alpha,
                         **kwargs)
        self.input_dim = input_dim
        self.generator_units_values = generator_units_values
        self.discriminator_units_values = discriminator_units_values
        self.generator_activation_hidden = generator_activation_hidden
        self.discriminator_activation_hidden = discriminator_activation_hidden
        self.generator_activation_out = generator_activation_out
        self.discriminator_activation_out = discriminator_activation_out
        self.num_discriminator_steps = num_discriminator_steps
        self.hint_rate = hint_rate
        self.alpha = alpha

    def get_config(self):
        config = {"name": self.name,
                  "trainable": self.trainable,
                  "input_dim": self.input_dim,
                  "generator_units_values": self.generator_units_values,
                  "discriminator_units_values": self.discriminator_units_values,
                  "generator_activation_hidden": self.generator_activation_hidden,
                  "discriminator_activation_hidden": self.discriminator_activation_hidden,
                  "generator_activation_out": self.generator_activation_out,
                  "discriminator_activation_out":  self.discriminator_activation_out,
                  "num_discriminator_steps": self.num_discriminator_steps,
                  "hint_rate": self.hint_rate,
                  "alpha": self.alpha,
                  }
        return config

    @classmethod
    def from_config(cls, config):
        input_dim = config.pop("input_dim")
        return cls(input_dim=input_dim,
                   **config)



from tensorflow import keras
from teras.utils.types import ActivationType


@keras.saving.register_keras_serializable(package="teras.layers.vime")
class VimeMaskEstimator(keras.layers.Layer):
    """
    VimeMaskEstimator layer based on the
    ``VIME`` architecture proposed by Jinsung Yoon et a.
    in the paper "VIME: Extending the Success of Self- and
    Semi-supervised Learning to Tabular Domain"

    Reference(s):
        https://proceedings.neurips.cc/paper/2020/hash/7d97667a3e056acab9aaf653807b4a03-Abstract.html

    Args:
        units: ``int``, default 32,
            Dimensionality of Mask Estimator layer

        activation: ``str`` or ``callable`` or ``keras.layers.Layer``, default "sigmoid",
            Activation function to use.
    """
    def __init__(self,
                 units: int = 32,
                 activation: ActivationType = "sigmoid",
                 **kwargs):
        super().__init__(**kwargs)
        self.units = units
        self.activation = activation
        self.estimator = keras.layers.Dense(self.units,
                                            activation=self.activation)

    def call(self, inputs):
        return self.estimator(inputs)

    def get_config(self):
        config = super().get_config()
        activation_serialized = self.activation
        if not isinstance(self.activation, str):
            activation_serialized = keras.layers.serialize(self.activation)
        config.update({'units': self.units,
                       'activation': activation_serialized,
                       }
                      )
        return config



import tensorflow as tf
from tensorflow import keras
from teras.models.node import NODEClassifier, NODERegressor
import os


class NODEClassifierTest(tf.test.TestCase):
    def test_valid_call(self):
        inputs = tf.ones((8, 5))
        model = NODEClassifier(num_classes=2,
                               input_dim=5)
        model(inputs)

    def test_save_and_load(self):
        inputs = tf.ones((8, 5))
        model = NODEClassifier(num_classes=2,
                               input_dim=5)
        save_path = os.path.join(self.get_temp_dir(), "node_classifier.keras")
        model.save(save_path, save_format="keras_v3")
        reloaded_model = keras.models.load_model(save_path)
        outputs_original = model(inputs)
        outputs_reloaded = reloaded_model(inputs)
        self.assertAllClose(outputs_original, outputs_reloaded)
        self.assertAllClose(model.weights, reloaded_model.weights)


class NODERegressorTest(tf.test.TestCase):
    def test_valid_call(self):
        inputs = tf.ones((8, 5))
        model = NODERegressor(num_outputs=2,
                              input_dim=5)
        model(inputs)

    def test_save_and_load(self):
        inputs = tf.ones((8, 5))
        model = NODERegressor(num_outputs=2,
                              input_dim=5)
        save_path = os.path.join(self.get_temp_dir(), "node_regressor.keras")
        model.save(save_path, save_format="keras_v3")
        reloaded_model = keras.models.load_model(save_path)
        outputs_original = model(inputs)
        outputs_reloaded = reloaded_model(inputs)
        self.assertAllClose(outputs_original, outputs_reloaded)
        self.assertAllClose(model.weights, reloaded_model.weights)



import tensorflow as tf
from teras.layers.tabnet.tabnet_decoder import TabNetDecoder


def test_tabnet_decoder_valid_call():
    decoder = TabNetDecoder(data_dim=5,
                            feature_transformer_dim=32,
                            decision_step_output_dim=32)
    inputs = tf.random.normal((8, 16))
    outputs = decoder(inputs)


def test_tabnet_decoder_output_shape():
    decoder = TabNetDecoder(data_dim=5,
                            feature_transformer_dim=32,
                            decision_step_output_dim=32)
    inputs = tf.random.normal((8, 16))
    outputs = decoder(inputs)
    assert len(tf.shape(outputs)) == 2
    assert tf.shape(outputs)[0] == 8
    assert tf.shape(outputs)[1] == 5



from tensorflow import keras
from teras.layers.dnfnet.dnnf import DNNF
from teras.utils.types import (IntegerSequence,
                               FloatSequence,
                               UnitsValuesType)
from teras.layers.common.head import (ClassificationHead,
                                      RegressionHead)
from teras.layerflow.models.dnfnet import DNFNet as _DNFNetLF


@keras.saving.register_keras_serializable(package="teras.models")
class DNFNet(_DNFNetLF):
    """
    DNFNet model based on the DNFNet architecture,
    proposed by Liran Katzir et al.
    in the paper,
    "NET-DNF: Effective Deep Modeling Of Tabular Data."

    Reference(s):
        https://openreview.net/forum?id=73WTGs96kho

    Args:
        input_dim: ``int``,
            Dimensionality of the input dataset,
            or the number of features in the dataset.

        num_dnnf_layers: ``int``, default 1,
            Number of ``DNNF`` layers to use in the model

        num_formulas: ``int``, default 256,
            Number of DNF formulas to use in each ``DNNF`` layer.
            Each DNF formula is analogous to a tree in tree based ensembles.

        num_conjunctions_arr: ``List[int]`` or ``Tuple[int]``, default [6, 9, 12, 15],
            Conjunctions array to use in each DNNF layer.
            It is used in the computation of total number of literals as well as
            computation of number of literals per DNF formula.

        conjunctions_depth_arr: ``List[int]`` or ``Tuple[int]``, default [2, 4, 6],
            Conjunctions depth array to use in each ``DNNF`` layer.
            It is used in the computation of total number of literals as well as
            computation of number of literals per DNF formula.

        keep_feature_prob_arr: ``List[float]`` or ``Tuple[float]``, default [0.1, 0.3, 0.5, 0.7, 0.9],
            Feature probability array to use in each ``DNNF`` layer.
            It is used by the Feature Selection layer to randomly select a probability
            value that is used in the random selection of input features.

        elastic_net_beta: ``float``, default 0.4,
            Used in the computation of Elastic Net Regularization in the ``DNNF`` layer.

        binary_threshold_eps: ``float``, default 1.0,
            Used in the computation of learnable mask in the ``DNNF`` layer.

        temperature: ``float``, default 2.0,
            Temperature value to use in the ``Localization`` layer.
            According to the paper, The inclusion of an adaptive temperature in this localization mechanism
            facilitates a data-dependent degree of exclusivity:
            at high temperatures, only a few DNNFs will handle an input instance whereas
            at low temperatures, more DNNFs will effectively participate in the ensemble.
    """
    def __init__(self,
                 input_dim: int,
                 num_dnnf_layers: int = 1,
                 num_formulas: int = 2048,
                 num_conjunctions_arr: IntegerSequence = [6, 9, 12, 15],
                 conjunctions_depth_arr: IntegerSequence = [2, 4, 6],
                 keep_feature_prob_arr: FloatSequence = [0.1, 0.3, 0.5, 0.7, 0.9],
                 elastic_net_beta: float = 0.4,
                 binary_threshold_eps: float = 1.0,
                 temperature: float = 2.0,
                 **kwargs):
        dnnf_layers = keras.models.Sequential(name="dnnf_layers")
        for _ in range(num_dnnf_layers):
            dnnf_layers.add(DNNF(num_formulas=num_formulas,
                                 num_conjunctions_arr=num_conjunctions_arr,
                                 conjunctions_depth_arr=conjunctions_depth_arr,
                                 keep_feature_prob_arr=keep_feature_prob_arr,
                                 elastic_net_beta=elastic_net_beta,
                                 binary_threshold_eps=binary_threshold_eps,
                                 temperature=temperature))
        super().__init__(input_dim=input_dim,
                         dnnf_layers=dnnf_layers,
                         **kwargs)
        self.input_dim = input_dim
        self.num_dnnf_layers = num_dnnf_layers
        self.num_formulas = num_formulas
        self.num_conjunctions_arr = num_conjunctions_arr
        self.conjunctions_depth_arr = conjunctions_depth_arr
        self.keep_feature_prob_arr = keep_feature_prob_arr
        self.elastic_net_beta = elastic_net_beta
        self.binary_threshold_eps = binary_threshold_eps
        self.temperature = temperature

    def get_config(self):
        config = {'name': self.name,
                  'trainable': self.trainable,
                  'input_dim': self.input_dim,
                  'num_dnnf_layers': self.num_dnnf_layers,
                  'num_formulas': self.num_formulas,
                  'num_conjunctions_arr': self.num_conjunctions_arr,
                  'conjunctions_depth_arr': self.conjunctions_depth_arr,
                  'keep_feature_prob_arr': self.keep_feature_prob_arr,
                  'elastic_net_beta': self.elastic_net_beta,
                  'binary_threshold_eps': self.binary_threshold_eps,
                  'temperature': self.temperature
                  }
        return config

    @classmethod
    def from_config(cls, config):
        input_dim = config.pop("input_dim")
        return cls(input_dim=input_dim,
                   **config)


@keras.saving.register_keras_serializable(package="teras.models")
class DNFNetClassifier(DNFNet):
    """
    DNFNetRegressor based on the DNFNet architecture proposed by Liran Katzir et al.
    in the paper NET-DNF: Effective Deep Modeling Of Tabular Data.

    Reference(s):
        https://openreview.net/forum?id=73WTGs96kho

    Args:
        num_classes: ``int``, default 2,
            Number of classes to predict.

        head_units_values: ``List[int]`` or ``Tuple[int]``, default None,
            Hidden units to use in the Classification head.
            For each value in the list/tuple,
            a hidden layer of that dimensionality is added to the head.
            By default, no hidden layer is used.

        input_dim: ``int``,
            Dimensionality of the input dataset,
            or the number of features in the dataset.

        num_dnnf_layers: ``int``, default 1,
            Number of ``DNNF`` layers to use in the model

        num_formulas: ``int``, default 256,
            Number of DNF formulas to use in each ``DNNF`` layer.
            Each DNF formula is analogous to a tree in tree based ensembles.

        num_conjunctions_arr: ``List[int]`` or ``Tuple[int]``, default [6, 9, 12, 15],
            Conjunctions array to use in each DNNF layer.
            It is used in the computation of total number of literals as well as
            computation of number of literals per DNF formula.

        conjunctions_depth_arr: ``List[int]`` or ``Tuple[int]``, default [2, 4, 6],
            Conjunctions depth array to use in each ``DNNF`` layer.
            It is used in the computation of total number of literals as well as
            computation of number of literals per DNF formula.

        keep_feature_prob_arr: ``List[float]`` or ``Tuple[float]``, default [0.1, 0.3, 0.5, 0.7, 0.9],
            Feature probability array to use in each ``DNNF`` layer.
            It is used by the Feature Selection layer to randomly select a probability
            value that is used in the random selection of input features.

        elastic_net_beta: ``float``, default 0.4,
            Used in the computation of Elastic Net Regularization in the ``DNNF`` layer.

        binary_threshold_eps: ``float``, default 1.0,
            Used in the computation of learnable mask in the ``DNNF`` layer.

        temperature: ``float``, default 2.0,
            Temperature value to use in the ``Localization`` layer.
            According to the paper, The inclusion of an adaptive temperature in this localization mechanism
            facilitates a data-dependent degree of exclusivity:
            at high temperatures, only a few DNNFs will handle an input instance whereas
            at low temperatures, more DNNFs will effectively participate in the ensemble.
    """
    def __init__(self,
                 num_classes: int = 2,
                 head_units_values: UnitsValuesType = None,
                 input_dim: int = None,
                 num_dnnf_layers: int = 1,
                 num_formulas: int = 2048,
                 num_conjunctions_arr: IntegerSequence = [6, 9, 12, 15],
                 conjunctions_depth_arr: IntegerSequence = [2, 4, 6],
                 keep_feature_prob_arr: FloatSequence = [0.1, 0.3, 0.5, 0.7, 0.9],
                 elastic_net_beta: float = 0.4,
                 binary_threshold_eps: float = 1.0,
                 temperature: float = 2.0,
                 **kwargs):
        head = ClassificationHead(num_classes=num_classes,
                                  units_values=head_units_values)
        super().__init__(input_dim=input_dim,
                         num_dnnf_layers=num_dnnf_layers,
                         num_conjunctions_arr=num_conjunctions_arr,
                         conjunctions_depth_arr=conjunctions_depth_arr,
                         keep_feature_prob_arr=keep_feature_prob_arr,
                         num_formulas=num_formulas,
                         elastic_net_beta=elastic_net_beta,
                         binary_threshold_eps=binary_threshold_eps,
                         temperature=temperature,
                         head=head,
                         **kwargs)
        self.num_classes = num_classes
        self.head_units_values = head_units_values

    def get_config(self):
        config = super().get_config()
        config.update({'num_classes': self.num_classes,
                       'head_units_values': self.head_units_values
                       })
        return config


@keras.saving.register_keras_serializable(package="teras.models")
class DNFNetRegressor(DNFNet):
    """
    DNFNetRegressor based on the DNFNet architecture proposed by Liran Katzir et al.
    in the paper NET-DNF: Effective Deep Modeling Of Tabular Data.

    Reference(s):
        https://openreview.net/forum?id=73WTGs96kho

    Args:
        num_outputs: ``int``, default 1,
            Number of regression outputs.

        head_units_values: ``List[int]`` or ``Tuple[int]``, None,
            Hidden units to use in the Classification head.
            For each value in the list/tuple,
            a hidden layer of that dimensionality is added to the head.
            By default, no hidden layer is used.

        input_dim: ``int``,
            Dimensionality of the input dataset,
            or the number of features in the dataset.

        num_dnnf_layers: ``int``, default 1,
            Number of ``DNNF`` layers to use in the model

        num_formulas: ``int``, default 256,
            Number of DNF formulas to use in each ``DNNF`` layer.
            Each DNF formula is analogous to a tree in tree based ensembles.

        num_conjunctions_arr: ``List[int]`` or ``Tuple[int]``, default [6, 9, 12, 15],
            Conjunctions array to use in each DNNF layer.
            It is used in the computation of total number of literals as well as
            computation of number of literals per DNF formula.

        conjunctions_depth_arr: ``List[int]`` or ``Tuple[int]``, default [2, 4, 6],
            Conjunctions depth array to use in each ``DNNF`` layer.
            It is used in the computation of total number of literals as well as
            computation of number of literals per DNF formula.

        keep_feature_prob_arr: ``List[float]`` or ``Tuple[float]``, default [0.1, 0.3, 0.5, 0.7, 0.9],
            Feature probability array to use in each ``DNNF`` layer.
            It is used by the Feature Selection layer to randomly select a probability
            value that is used in the random selection of input features.

        elastic_net_beta: ``float``, default 0.4,
            Used in the computation of Elastic Net Regularization in the ``DNNF`` layer.

        binary_threshold_eps: ``float``, default 1.0,
            Used in the computation of learnable mask in the ``DNNF`` layer.

        temperature: ``float``, default 2.0,
            Temperature value to use in the ``Localization`` layer.
            According to the paper, The inclusion of an adaptive temperature in this localization mechanism
            facilitates a data-dependent degree of exclusivity:
            at high temperatures, only a few DNNFs will handle an input instance whereas
            at low temperatures, more DNNFs will effectively participate in the ensemble.
    """
    def __init__(self,
                 num_outputs: int = 1,
                 head_units_values: UnitsValuesType = None,
                 input_dim: int = None,
                 num_dnnf_layers: int = 1,
                 num_formulas: int = 2048,
                 num_conjunctions_arr: IntegerSequence = [6, 9, 12, 15],
                 conjunctions_depth_arr: IntegerSequence = [2, 4, 6],
                 keep_feature_prob_arr: FloatSequence = [0.1, 0.3, 0.5, 0.7, 0.9],
                 elastic_net_beta: float = 0.4,
                 binary_threshold_eps: float = 1.0,
                 temperature: float = 2.0,
                 **kwargs):
        head = RegressionHead(num_outputs=num_outputs,
                              units_values=head_units_values)
        super().__init__(input_dim=input_dim,
                         num_dnnf_layers=num_dnnf_layers,
                         num_conjunctions_arr=num_conjunctions_arr,
                         conjunctions_depth_arr=conjunctions_depth_arr,
                         keep_feature_prob_arr=keep_feature_prob_arr,
                         num_formulas=num_formulas,
                         elastic_net_beta=elastic_net_beta,
                         binary_threshold_eps=binary_threshold_eps,
                         temperature=temperature,
                         head=head,
                         **kwargs)
        self.num_outputs = num_outputs
        self.head_units_values = head_units_values

    def get_config(self):
        config = super().get_config()
        new_config = {'num_outputs': self.num_outputs,
                      'head_units_values': self.head_units_values
                      }
        config.update(new_config)
        return config
